import os
from urllib.parse import quote
from datetime import datetime, timedelta
from google.cloud import storage 
import google.auth
from google.auth.transport import requests

from ..logging import log


def sign_gcs_url(bucket_name:str, object_name:str, expiry_secs = 86400):
    # https://stackoverflow.com/questions/64234214/how-to-generate-a-blob-signed-url-in-google-cloud-run

    credentials, project_id = google.auth.default()

    # Perform a refresh request to get the access token of the current credentials (Else, it's None)
    r = requests.Request()
    credentials.refresh(r)

    client = storage.Client()
    bucket = client.get_bucket(bucket_name)
    blob = bucket.get_blob(object_name)
    if not blob:
        return None

    expires = datetime.now() + timedelta(seconds=expiry_secs)

    # If you use a service account credential, you can use the embedded email
    if hasattr(credentials, "service_account_email"):
        service_account_email = credentials.service_account_email
    else:
        service_account_email = os.getenv('GCS_MAIL_USER')
        if service_account_email is None:
            log.error("For local testing must set a GCS_MAIL_USER to sign GCS URLs")
        log.error("Could not create the credentials for signed requests - no credentials.service_account_email or GCS_MAIL_USER with roles/iam.serviceAccountTokenCreator")
        return None

    url = blob.generate_signed_url(
        version="v4",
        expiration=expires,
        service_account_email=service_account_email, 
        access_token=credentials.token)
    log.debug(f"Generated signed URL: {url}")
    return url


def construct_download_link(source_uri: str) -> str:
    """Creates a viewable Cloud Storage web browser link from a gs:// URI.""" 
    if not source_uri.startswith("gs://"):
        return source_uri, source_uri, False  # Return the URI as is if it doesn't start with gs://

    bucket_name, object_name = parse_gs_uri(source_uri)

    signed_url = sign_gcs_url(bucket_name, object_name)
    if signed_url:
        return signed_url, os.path.basename(object_name), True
    
    log.error(f"Failed to generate signed URL for {source_uri}")
    return construct_download_link_simple(bucket_name, object_name)


def construct_download_link_simple(bucket_name:str, object_name:str) -> str:
    """Creates a viewable Cloud Storage web browser link from a gs:// URI.

    Args:
        source_uri: The gs:// URI of the object in Cloud Storage.

    Returns:
        A URL that directly access the object in the Cloud Storage web browser.
    """

    public_url = f"https://storage.cloud.google.com/{bucket_name}/{quote(object_name)}"
    filename = os.path.basename(object_name)
    signed = False
    return public_url, filename, signed

def parse_gs_uri(gs_uri: str) -> tuple[str, str]:
    """Parses a gs:// URI into the bucket name and object name.

    Args:
        gs_uri: The gs:// URI to parse.

    Returns:
        A tuple containing the bucket name and object name.
    """
    parts = gs_uri.split("/")
    if len(parts) < 3:
        raise ValueError(f"Invalid gs:// URI: {gs_uri}")
    return parts[2], "/".join(parts[3:])


#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from ..logging import log
from ..utils.parsers import remove_whitespace
from langchain.schema import Document
import langchain.text_splitter as text_splitter
from .images import upload_doc_images
from .doc_handling import send_doc_to_docstore, summarise_docs


def chunk_doc_to_docs(documents: list, extension: str = ".md", min_size: int = 800, vector_name=None, **kwargs):
    """Turns a Document object into a list of many Document chunks.
       If a document or chunk is smaller than min_size, it will be merged with adjacent documents or chunks."""

    if documents is None:
        return None

    # send full parsed doc to docstore
    send_doc_to_docstore(documents, vector_name=vector_name)

    documents = summarise_docs(documents, vector_name=vector_name)

    # Combine entire documents that are smaller than min_size
    combined_documents_content = ""
    combined_documents = []
    for document in documents:
        content = remove_whitespace(document.page_content)

        # look for images and upload them for later extraction, add metadata of location
        image_gsurl = upload_doc_images(document.metadata)
        if image_gsurl:
            document.metadata["image_gsurl"] = image_gsurl

        if len(content) < min_size:
            combined_documents_content += content + "\n"
            log.info(f"Appending document as its smaller than {min_size}: length {len(content)} - appended doc length {len(combined_documents_content)}")
        else:
            if combined_documents_content:
                combined_documents.append(Document(page_content=combined_documents_content, metadata=document.metadata))
                combined_documents_content = ""
            combined_documents.append(document)

    if combined_documents_content:
        combined_documents.append(Document(page_content=combined_documents_content, metadata=documents[-1].metadata))

    source_chunks = []
    temporary_chunk = ""
    chunk_number = 0
    for document in combined_documents:
        splitter = choose_splitter(extension, vector_name=vector_name, **kwargs)
        for chunk in splitter.split_text(document.page_content):
            # If a chunk is smaller than the min_size, append it to temporary_chunk with a line break and continue
            if len(chunk) < min_size:
                temporary_chunk += chunk + "\n"
                log.info(f"Appending chunk as its smaller than {min_size}: length {len(chunk)}")
                continue

            # If there's content in temporary_chunk, append it to the current chunk
            if temporary_chunk:
                chunk = temporary_chunk + chunk
                temporary_chunk = ""

            # If the combined chunk is still less than the min_size, append to temporary_chunk with a line break and continue
            if len(chunk) < min_size:
                temporary_chunk += chunk + "\n"
                log.info(f"Appending chunk as its smaller than {min_size}: length {len(chunk)}")
                continue
            
            log.info(f"Adding chunk of length {len(chunk)}")
            document.metadata["chunk_number"] = chunk_number
            source_chunks.append(Document(page_content=chunk, metadata=document.metadata))

        # If there's any remaining content in temporary_chunk, append it as a new chunk
        if temporary_chunk:
            source_chunks.append(Document(page_content=temporary_chunk, metadata=document.metadata))
            temporary_chunk = ""
        
        chunk_number += 1

    log.info(f"Chunked into {chunk_number} documents")
    return source_chunks

def choose_splitter(extension: str, chunk_size: int=1024, chunk_overlap:int=0, vector_name: str=None):

    if vector_name:
        # check if there is a chunking configuration
        from ..utils import load_config_key
        chunk_config = load_config_key("chunker", vector_name=vector_name, filename="config/llm_config.yaml")
        if chunk_config:
            if chunk_config.get("type") == "semantic":
                embedding_str = chunk_config.get("llm")
                if not embedding_str:
                    log.error("Unable to find embedding 'config.chunker.llm' configuration needed for semantic chunking")
                else:
                    log.info(f"Semantic chunking for {vector_name}")
                    from langchain_experimental.text_splitter import SemanticChunker
                    from ..components import pick_embedding
                    embeddings = pick_embedding(embedding_str)
                    semantic_splitter = SemanticChunker(
                        embeddings, breakpoint_threshold_type="percentile"
                    )

                    return semantic_splitter


    if extension == ".py":
        return text_splitter.PythonCodeTextSplitter()
    elif extension == ".md":
        return text_splitter.MarkdownTextSplitter()
    elif extension == ".html":
        return text_splitter.HTMLHeaderTextSplitter()
    
    return text_splitter.RecursiveCharacterTextSplitter(chunk_size=chunk_size, chunk_overlap=chunk_overlap)



import os
from ..logging import log_folder_location

def get_db_directory(db_dir='db'):
    current_script_directory = os.path.dirname(os.path.abspath(__file__))
    # Get the parent directory of the current script directory
    parent_directory = os.path.abspath(os.path.join(current_script_directory, os.pardir))
    db_directory = os.path.join(parent_directory, db_dir)
    log_folder_location(db_directory)

    return db_directory